﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.ComponentModel;

namespace WpfApplication1
{
    public enum BindableListAddMode { End, Begin }

    public sealed class BindableList<T> : IEnumerable<T>, IBindingList, IDisposable
    {
        #region Fields

        //private List<PropertyDescriptor> _descriptors = new List<PropertyDescriptor>();

        private List<T> _items = new List<T>();

        #endregion

        #region Constructors

        public BindableList() { }

        public BindableList(IEnumerable<T> collection)
        {
            if (collection != null)
                _items.AddRange(collection);
        }

        #endregion

        #region IBindingList Members

        public void AddIndex(PropertyDescriptor property)
        {
            //if (!_descriptors.Contains(property))
            //    _descriptors.Add(property);
        }

        public object AddNew()
        {
            T item = Activator.CreateInstance<T>();
            lock (_items)
            {
                if (AddMode == BindableListAddMode.End)
                    _items.Add(item);
                else
                    _items.Insert(0, item);
                AttachItem(item as INotifyPropertyChanged);
                if (this.ListChanged != null)
                {
                    this.ListChanged(this, new ListChangedEventArgs(ListChangedType.ItemAdded, AddMode == BindableListAddMode.Begin ? 0 : _items.Count - 1));
                }
            }
            return item;
        }

        public bool AllowEdit { get; set; }

        public bool AllowNew { get; set; }

        public bool AllowRemove { get; set; }

        public void ApplySort(PropertyDescriptor property, ListSortDirection direction)
        {

        }

        public int Find(PropertyDescriptor property, object key)
        {
            return -1;
        }

        public bool IsSorted
        {
            get { return false; }
        }

        public event ListChangedEventHandler ListChanged;

        public void RemoveIndex(PropertyDescriptor property)
        {
            //_descriptors.Remove(property);
        }

        public void RemoveSort()
        {

        }

        public ListSortDirection SortDirection { get; set; }

        public PropertyDescriptor SortProperty { get; set; }

        public bool SupportsChangeNotification { get { return true; } }

        public bool SupportsSearching { get { return false; } }

        public bool SupportsSorting { get { return false; } }

        #endregion

        #region IList Members

        public int Add(T value)
        {
            return Add((object)value);
        }

        public int Add(object value)
        {
            lock (_items)
            {
                if (value is T)
                {
                    if (AddMode == BindableListAddMode.End)
                        _items.Add((T)value);
                    else
                        _items.Insert(0, (T)value);
                    AttachItem(value as INotifyPropertyChanged);
                    if (this.ListChanged != null)
                    {
                        this.ListChanged(this, new ListChangedEventArgs(ListChangedType.ItemAdded, AddMode == BindableListAddMode.Begin ? 0 : _items.Count - 1));
                    }
                    return _items.Count - 1;
                }
            }
            return -1;
        }

        public void Clear()
        {
            lock (_items)
            {
                _items.ForEach(i => DetachItem(i as INotifyPropertyChanged));
                _items.Clear();
                if (this.ListChanged != null)
                {
                    this.ListChanged(this, new ListChangedEventArgs(ListChangedType.Reset, -1));
                }
            }
        }

        public bool Contains(object value)
        {
            lock (_items)
            {
                if (value is T)
                    return _items.Contains((T)value);
            }
            return false;
        }

        public int IndexOf(object value)
        {
            lock (_items)
            {
                if (value is T)
                    return _items.IndexOf((T)value);
            }
            return -1;
        }

        public void Insert(int index, object value)
        {
            lock (_items)
            {
                if (value is T)
                {
                    _items.Insert(index, (T)value);
                    AttachItem(value as INotifyPropertyChanged);
                    if (this.ListChanged != null)
                    {
                        this.ListChanged(this, new ListChangedEventArgs(ListChangedType.ItemAdded, index));
                    }
                }
            }
        }

        public bool IsFixedSize { get { return false; } }

        public bool IsReadOnly
        {
            get { return !AllowEdit && !AllowNew && !AllowRemove; }
        }

        private void RemoveInternalAt(int index)
        {
            lock (_items)
            {
                if (index >= 0 && index < _items.Count)
                {
                    if (this.ListChanged != null)
                    {
                        this.ListChanged(this, new ListChangedEventArgs(ListChangedType.ItemDeleted, index));
                    }
                    DetachItem(_items[index] as INotifyPropertyChanged);
                    _items.RemoveAt(index);
                }
            }
        }

        public void Remove(object value)
        {
            RemoveInternalAt(_items.FindIndex(i => object.Equals(i, value)));
        }

        public void RemoveAt(int index)
        {
            RemoveInternalAt(index);
        }

        public object this[int index]
        {
            get
            {
                lock (_items)
                {
                    return _items[index];
                }
            }
            set
            {
                lock (_items)
                {
                    if (value is T)
                    {
                        _items[index] = (T)value;
                    }
                }
            }
        }

        #endregion

        public T Get(int index)
        {
            return (T)_items[index];
        }

        #region ICollection Members

        public void CopyTo(Array array, int index)
        {
            lock (_items)
            {
                T[] arrayCast = array as T[];
                if (arrayCast != null)
                {
                    _items.CopyTo(arrayCast, index);
                }
            }
        }

        public int Count
        {
            get
            {
                lock (_items)
                {
                    return _items.Count;
                }
            }
        }

        public bool IsSynchronized
        {
            get { return true; }
        }

        public object SyncRoot
        {
            get { return null; }
        }

        #endregion

        #region IEnumerable Members

        public System.Collections.IEnumerator GetEnumerator()
        {
            lock (_items)
            {
                foreach (object item in _items)
                    yield return item;
            }
        }

        #endregion

        public BindableListAddMode AddMode { get; set; }

        #region IDisposable Members

        public void Dispose()
        {
            //_descriptors.Clear();
            //_descriptors = null;
            Clear();
            _items = null;
        }

        #endregion

        #region IEnumerable<T> Members

        IEnumerator<T> IEnumerable<T>.GetEnumerator()
        {
            lock (_items)
            {
                foreach (var item in _items)
                    yield return item;
            }
        }

        #endregion

        #region Custom Methods

        private void AttachItem(INotifyPropertyChanged item)
        {
            if(item != null)
                item.PropertyChanged += ItemPropertyChanged;
        }

        private void ItemPropertyChanged(object sender, PropertyChangedEventArgs e)
        {
            if(this.ListChanged != null)
                this.ListChanged(this, new ListChangedEventArgs(ListChangedType.ItemChanged, _items.IndexOf((T)sender)));
        }

        private void DetachItem(INotifyPropertyChanged item)
        {
            if (item != null)
                item.PropertyChanged -= ItemPropertyChanged;
        }

        #endregion
    }
}
